#!/usr/bin/env python
#-*- coding:utf-8 -*-
#############################################
# File Name: setup.py
# Author:
# Mail:
# Created Time:
#############################################
from setuptools import setup, find_packages
setup(
    name = "TRDstest",
    version = "0.1.7",
    keywords = ("pip", "tensor","tensor ring decomposition", "", ""),
    description = "tensor ring decomposition toolbox",
    long_description = "tensor ring decomposition toolbox",
    license = "BSD 3-Clause License",
    url = "",
    author = "",
    author_email = "",
    packages = find_packages(),
    include_package_data = True,
    platforms = "any",
    install_requires = []
)